/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001-2004  by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@gmx.de
 *    Borg Enders  borg@borgsoft.de
 *
 *********************************************************************/

#include "cards_information.h"

#include "constants.h"

#include "../../card/trick.h"

// whether to make self checks
// local macro
#ifndef RELEASE
#define SELF_CHECK
#endif

/**
 ** constructor
 **
 ** @param	player			corresponding player
 ** @param	remembertrickno		how many tricks are remembered
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
CardsInformation::CardsInformation(Player const& player,
				   unsigned const remembertrickno) :
  game_(&player.game()),
  player_(&player),
  trick_(NULL),
  remembertrickno_(remembertrickno),
  of_player_(),
  played_()
{
  for (unsigned p = 0; p < this->game().playerno(); ++p)
    this->of_player_.push_back(OfPlayer(*this, p));
}

/**
 ** constructor
 **
 ** @param	game			corresponding game
 ** @param	remembertrickno		how many tricks are remembered
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
CardsInformation::CardsInformation(Game const& game,
				   unsigned const remembertrickno) :
  game_(&game),
  player_(NULL),
  trick_(NULL),
  remembertrickno_(remembertrickno),
  of_player_(),
  played_()
{
  for (unsigned p = 0; p < this->game().playerno(); ++p)
    this->of_player_.push_back(OfPlayer(*this, p));
}

/**
 ** copy constructor
 **
 ** @param	cards_information	object to be copied
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
CardsInformation::CardsInformation(CardsInformation const& cards_information) :
  game_(cards_information.game_),
  player_(cards_information.player_),
  trick_(cards_information.trick_),
  remembertrickno_(cards_information.remembertrickno_),
  of_player_(cards_information.of_player_),
  played_(cards_information.played_)
{
  for (vector<OfPlayer>::iterator p = this->of_player_.begin();
       p != this->of_player_.end();
       ++p)
    p->set_cards_information(*this);

#ifdef SELF_CHECK
  this->self_check();
#endif
} // CardsInformation::CardsInformation(CardsInformation cards_information)

/**
 ** copy operator
 **
 ** @param	cards_information	object to be copied
 **
 ** @return	object
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
CardsInformation&
CardsInformation::operator=(CardsInformation const& cards_information)
{
  this->game_ = cards_information.game_;
  this->player_ = cards_information.player_;
  this->trick_ = cards_information.trick_;
  this->remembertrickno_ = cards_information.remembertrickno_;
  this->of_player_ = cards_information.of_player_;
  for (vector<OfPlayer>::iterator p = this->of_player_.begin();
       p != this->of_player_.end();
       ++p)
    p->set_cards_information(*this);

  this->played_ = cards_information.played_;

#ifdef SELF_CHECK
  this->self_check();
#endif

  return *this;
} // CardsInformation& CardsInformation::operator=(CardsInformation cards_information)

/**
 ** destructor
 **
 ** @param	-
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
CardsInformation::~CardsInformation()
{ }

/**
 ** writes 'cards_information' in 'ostr'
 **
 ** @param	ostr			output stream
 ** @param	cards_information	object to write into 'ostr'
 **
 ** @return	the output stream
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
ostream&
operator<<(ostream& ostr, CardsInformation const& cards_information)
{
  cards_information.write(ostr);
  return ostr;
} // ostream& operator<<(ostream& ostr, CardsInformation cards_information)

/**
 ** writes 'cards_information' in 'ostr'
 **
 ** @param	ostr		output stream
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::write(ostream& ostr) const
{
  ostr << "cards information:\n"
    << "{\n";

  ostr << "played:\n"
    << "{\n"
    << this->played_
    << "}\n"
    << '\n';

  for (vector<OfPlayer>::const_iterator p = this->of_player_.begin();
       p != this->of_player_.end();
       ++p)
    ostr << *p << '\n';

  ostr << "}\n";

  return ;
} // void CardsInformation::write(ostream& ostr) const

/**
 ** resets all information
 **
 ** @param	-
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::reset()
{
  this->played_.clear();

  for (vector<OfPlayer>::iterator p = this->of_player_.begin();
       p != this->of_player_.end();
       ++p)
    p->reset();

  return ;
} // void CardsInformation::reset()

/**
 ** recalcs all information
 ** takes 'Aiconfig::REMEMBERTRICKS' into account
 **
 ** @param	-
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::recalc()
{
  this->reset();

  if (this->player_)
    this->set_hand(this->player(), this->player().hand());

  unsigned t = 0;
  if (this->game().trick_current_no() > this->remembertrickno())
    t = this->game().trick_current_no() - this->remembertrickno();

  for (t = t; t < this->game().trick_current_no(); ++t) {
    this->trick_ = &this->game().trick(t);
    for (unsigned c = 0; c < this->trick().actcardno(); ++c)
      this->card_played(this->trick().card(c));
  }

  return ;
} // void CardsInformation::recalc()

/**
 ** -> result
 **
 ** @param	player	the player whose information is asked for
 **
 ** @return	the information of 'player'
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
CardsInformation::OfPlayer const&
CardsInformation::of_player(Player const& player) const
{
  return this->of_player(player.no());
} // CardsInformation::OfPlayer const& CardsInformation::of_player(Player player) const

/**
 ** -> result
 **
 ** @param	playerno	the player whose information is asked for
 **
 ** @return	the information of player 'playerno'
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
CardsInformation::OfPlayer const&
CardsInformation::of_player(unsigned const playerno) const
{
  return this->of_player_[playerno];
} // CardsInformation::OfPlayer CardsInformation::of_player(unsigned playerno) const

/**
 ** -> result
 **
 ** @param	player	the player whose hand is asked for
 **
 ** @return	hand with all cards the player can have
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
Hand
CardsInformation::possible_hand(Player const& player) const
{
  DEBUG_ASSERTION((&this->game() == &player.game()),
		  "CardsInformation::hand(player):\n"
		  "  the game of the player is not the game of the player");

  return this->of_player_[player.no()].possible_hand();
} // Hand CardsInformation::possible_hand(Player player) const

/**
 ** sets (changes) the game
 **
 ** @param	game	new game
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::set_game(Game const& game)
{
  this->game_ = &game;
  if (this->player_)
    this->player_ = &game.player(this->player().no());
  if (this->trick_)
    this->trick_ = &game.trick(this->trick().no());

  return ;
} // void CardsInformation::set_game(Game game)

/**
 ** sets the hand of the player
 **
 ** @param	player	player whose hand is set
 ** @param	hand	the hand of the player
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::set_hand(Player const& player, Hand const& hand)
{
  if (::game_status == GAMESTATUS::GAME_POVERTY_SHIFT)
    return ;
  this->of_player_[player.no()].add_can_only_have(hand.cards());

#ifdef SELF_CHECK
  this->self_check();
#endif

  return ;
} // void CardsInformation::set_hand(Player player, Hand hand)

/**
 ** updates the information of the card
 **
 ** @param	card	card to update the information of
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::update_information(Card const& card)
{
  for (vector<OfPlayer>::iterator p = this->of_player_.begin();
       p != this->of_player_.end();
       ++p)
    p->update_information(card);

  return ;
} // void CardsInformation::update_information(Card card)

/**
 ** -> result
 **
 ** @param	-
 **
 ** @return	how many cards have been played in total in the game
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
unsigned
CardsInformation::played_cards_no() const
{
  return this->played_.cards_no();
} // unsigned CardsInformation::played_cards_no() const

/**
 ** -> result
 **
 ** @param	card	card to look at
 **
 ** @return	how many of 'card' have been played
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
unsigned
CardsInformation::played(Card const& card) const
{
  return (this->played_)[card];
} // unsigned CardsInformation::played(Card card) const

/**
 ** the game starts
 **
 ** @param	-
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::game_start()
{
  switch (this->game().type()) {
  case GAMETYPE::MARRIAGE:
    this->of_player_[this->game().soloplayer().no()].add_must_have(Card::CLUB_QUEEN,
								   2);
    break;
  case GAMETYPE::POVERTY:
    if (this->game().poverty_cardno() == 0)
      this->of_player_[this->game().soloplayer().no()].add_does_not_have(Card::TRUMP);
    break;
  default:
    break;
  } // switch (this->game().type())

  return ;
} // void CardsInformation::game_start()

/**
 **
 ** the trick is opened
 **
 ** @param      trick	opened trick
 **
 ** @return     -
 **
 ** @author     Diether Knof
 **
 ** @version    0.6.9
 **
 **/
void
CardsInformation::trick_open(Trick const& trick)
{ 
  this->trick_ = &trick;
  if (this->game().trick_current_no() > this->remembertrickno())
    this->recalc();
  this->trick_ = &trick;

  return ;
} // void CardsInformation::trick_open(Trick const& trick)

/**
 ** 'card' has been played - update the information
 **
 ** @param	card	played card
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::card_played(HandCard const& card)
{
  this->played_.inc(card);
  DEBUG_ASSERTION(this->played(card)
		  <= this->game().rule()(Rule::NUMBER_OF_SAME_CARDS),
		  "CardsInformation::card_played(" << card << ")\n"
		  "  card is played more often than it is in the game");

  this->of_player_[card.player().no()].card_played(card);

  if (card.tcolor() != this->trick().startcard().tcolor()) {
    this->of_player_[card.player().no()].add_does_not_have(this->trick().startcard().tcolor());
  }

  // check whether the player has played as many trump as he has got back
  if ((this->game().type() == GAMETYPE::POVERTY)
      && (card.player().no() == this->game().soloplayer().no())
      && !this->of_player(this->game().soloplayer().no()
			 ).does_not_have(Card::TRUMP)
      && card.istrump()
      && (this->of_player(this->game().soloplayer().no()).played(Card::TRUMP)
	  == this->game().poverty_cardno()))
    this->of_player_[this->game().soloplayer().no()].add_does_not_have(Card::TRUMP);

  this->update_information(card);

#ifdef SELF_CHECK
  this->self_check();
#endif

  return ;
} // void CardsInformation::card_played(HandCard card)

/**
 ** an announcement has been made - update the information
 **
 ** @param	announcement	made announcement
 ** @param	player		player who has made the announcement
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::announcement_made(Announcement const& announcement,
				    Player const& player)
{
  if (this->game().type() == GAMETYPE::NORMAL) {

    try {
      if (this->game().teaminfo(player) == TEAM::RE) {
	if (this->of_player_[player.no()].played(Card::CLUB_QUEEN) == 0)
	  this->of_player_[player.no()].add_must_have(Card::CLUB_QUEEN);
      } else if (this->game().teaminfo(player) == TEAM::CONTRA) {
	this->of_player_[player.no()].add_cannot_have(Card::CLUB_QUEEN);
      } else {
	DEBUG_ASSERTION(false,
			"CardsInformation::announcement_made("
			<< announcement << ", " << player.no() << "):\n"
			"  team of the player not known in the game: "
			<< this->game().teaminfo(player));
      }

      this->update_information(Card::CLUB_QUEEN);
#ifdef WORKAROUND
      // In a virtual game it can be that there are more than two re-players.
      // Here we just allow more than two 'Re' players, t.i. players with
      // a club queen.
    } catch (...) {
      if (!this->game().isvirtual())
	throw;
#endif
    }
  } // if (this->game().type() == GAMETYPE::NORMAL)

#ifdef SELF_CHECK
  this->self_check();
#endif

  return ;
} // void CardsInformation::announcement_made(Announcement announcement, Player player)

/**
 ** a player has announced swines - update the information
 **
 ** @param	player	player who has swines
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::swines_announced(Player const& player)
{
  Card const swine(this->game().trumpcolor(), Card::ACE);
  this->of_player_[player.no()].add_must_have(swine, 2);
  for (vector<OfPlayer>::iterator p = this->of_player_.begin();
       p != this->of_player_.end();
       ++p)
    if (p->playerno() != player.no())
      this->of_player_[p->playerno()].add_cannot_have(swine);

#ifdef SELF_CHECK
  this->self_check();
#endif

  return ;
} // void CardsInformation::swines_announced(Player player)

/**
 ** a player has announced hyperswines - update the information
 **
 ** @param	player	player who has hyperswines
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::hyperswines_announced(Player const& player)
{
  Card const hyperswine(this->game().trumpcolor(), Card::NINE);
  this->of_player_[player.no()].add_must_have(hyperswine, 2);
  for (vector<OfPlayer>::iterator p = this->of_player_.begin();
       p != this->of_player_.end();
       ++p)
    if (p->playerno() != player.no())
      this->of_player_[p->playerno()].add_cannot_have(hyperswine);

#ifdef SELF_CHECK
  this->self_check();
#endif

  return ;
} // void CardsInformation::hyperswines_announced(Player player)

/**
 ** player 'genscher' has a genscher
 **
 ** @param	genscher	player with the genscher
 ** @param	partner		selected partner
 **
 ** @return	-
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
void
CardsInformation::genscher(Player const& genscher, Player const& partner)
{
  Card const genscher_card(this->game().trumpcolor(), Card::KING);
  this->of_player_[genscher.no()].add_must_have(genscher_card, 2);
  for (vector<OfPlayer>::iterator p = this->of_player_.begin();
       p != this->of_player_.end();
       ++p)
    if (p->playerno() != genscher.no())
      this->of_player_[p->playerno()].add_cannot_have(genscher_card);

#ifdef SELF_CHECK
  this->self_check();
#endif

  return ;
} // void CardsInformation::genscher(Player genscher, Player partner)

/**
 ** self check
 ** when an error is found, an ASSERTION is created
 **
 ** @param	-
 **
 ** @return	whether the self-check was successful (no error)
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
bool
CardsInformation::self_check() const
{
  if (this->played_.cards_no() > this->game().played_cards_no()) {
    DEBUG_ASSERTION(false,
		    "CardsInformation::self_check():\n"
		    "  played cards no = " << this->played_.cards_no()
		    << " > " << this->game().played_cards_no()
		    << " = game.played_cards_no()");
    return false;
  }
  for (CardCounter::const_iterator p = this->played_.begin();
       p != this->played_.end();
       ++p) {
    if (p->second > this->game().rule()(Rule::NUMBER_OF_SAME_CARDS)) {
      DEBUG_ASSERTION(false,
		      "CardsInformation::self_check():\n"
		      "  card '" << p->first << "' has been played"
		      << p->second << "times.\n"
		     );
      return false;
    }
    unsigned n = 0;
    for (vector<OfPlayer>::const_iterator f = this->of_player_.begin();
	 f != this->of_player_.end();
	 ++f)
      n += f->played(p->first);

    if (p->second != n) {
      DEBUG_ASSERTION(false,
		      "CardsInformation::self_check():\n"
		      "  card '" << p->first << "' has been played"
		      << p->second << "times, but the players say "
		      << n << " times.\n"
		     );
      return false;
    }
  } // for (p \in this->played_)

  for (vector<OfPlayer>::const_iterator f = this->of_player_.begin();
       f != this->of_player_.end();
       ++f)
    if (!f->self_check())
      return false;

  return true;
} // bool CardsInformation::self_check() const

/**
 ** -> result
 **
 ** @param	cards_information_a	first object
 ** @param	cards_information_b	second object
 **
 ** @return	whether the two objects are equal (the game may differ)
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
bool
operator==(CardsInformation const& cards_information_a,
	   CardsInformation const& cards_information_b)
{
  return ( (cards_information_a.remembertrickno_
	    == cards_information_b.remembertrickno_)
	   && (cards_information_a.of_player_
	       == cards_information_b.of_player_)
	   && (cards_information_a.played_
	       == cards_information_b.played_) );
} // bool operator==(CardsInformation cards_information_a, CardsInformation cards_information_b)

/**
 ** -> result
 **
 ** @param	cards_information_a	first object
 ** @param	cards_information_b	second object
 **
 ** @return	whether the two objects are different
 **
 ** @author	Diether Knof
 **
 ** @version	0.6.9
 **/
bool
operator!=(CardsInformation const& cards_information_a,
	   CardsInformation const& cards_information_b)
{
  return !(cards_information_a == cards_information_b);
} // bool operator!=(CardsInformation cards_information_a, CardsInformation cards_information_b)

#ifdef SELF_CHECK
#undef SELF_CHECK
#endif
